import threading
from abc import ABCMeta
from logging import Logger
from typing import Optional

from jsonpickle import json
from retrying import retry

from sidecar.aws_error_helper import AwsErrorHelper
from sidecar.aws_session import AwsSession
from sidecar.azure_clp.azure_clients import AzureClientsManager
from sidecar.const import Const
from sidecar.model.objects import AzureSidecarConfiguration, ISidecarConfiguration, AwsSidecarConfiguration
from sidecar.services.exceptions import NotReadyYetError
from sidecar.utils import Utils


class SandboxMetadataFetcher(metaclass=ABCMeta):
    def get_value(self) -> str:
        raise NotImplementedError()


class SandboxPublicAddressFetcher(SandboxMetadataFetcher):
    ATTEMPT_TIMEOUT_MINUTES = 30
    ATTEMPT_INTERVAL_SECONDS = 10
    NOT_AVAILABLE = 'NotAvailable'

    def __init__(self, logger: Logger, config: ISidecarConfiguration):
        self._sandbox_id = config.sandbox_id
        self._config = config
        self._logger = logger
        self._value = None
        self._thread = None
        self._lock = threading.RLock()

    def get_value(self) -> str:
        if not self._config.ingress_enabled:
            return self.NOT_AVAILABLE

        with self._lock:
            if not self._value and not self._thread:
                result = self._one_time_safe_fetch()
                if result:
                    self._value = result
                else:
                    self._thread = threading.Thread(target=self._fetching_loop, daemon=True)
                    self._thread.start()
            if self._value:
                return self._value
            raise NotReadyYetError()

    def _fetching_loop(self):
        self._logger.info(f'Staring trying to get sandbox public address')
        result = Utils.retry_on_exception(
            func=lambda: self._get_sandbox_public_ip(),
            interval_in_sec=self.ATTEMPT_INTERVAL_SECONDS,
            timeout_in_sec=self.ATTEMPT_TIMEOUT_MINUTES * 60,
            logger=self._logger,
            logger_msg='getting sandbox public address',
            log_every_n_attempts=6 * 2)  # 12 attempts = ~2min
        with self._lock:
            self._value = result
        self._logger.info(f'Sandbox public address is "{result}"')

    def _one_time_safe_fetch(self) -> Optional[str]:
        try:
            if not self._config.ingress_enabled:
                return self.NOT_AVAILABLE
            return self._get_sandbox_public_ip()
        except:
            return None

    def _get_sandbox_public_ip(self) -> str:
        raise NotImplementedError()


class AwsSandboxPublicAddressFetcher(SandboxPublicAddressFetcher):
    def __init__(self, logger: Logger, config: AwsSidecarConfiguration, aws_session: AwsSession):
        super().__init__(logger=logger, config=config)
        self._session = aws_session

    # Exponential Backoff starts from 1 sec to 2 min, until 2 min, in case of throttling
    @retry(wait_exponential_multiplier=1000,
           wait_exponential_max=1000 * 60 * 2,
           stop_max_delay=1000 * 60 * 2,
           retry_on_exception=AwsErrorHelper.is_throttling_error)
    def _get_sandbox_public_ip(self) -> str:
        self._logger.info(f'**** _get_sandbox_public_ip: get stack {self._session.infra_stack_name}')

        stack = self._session.get_cf_resource().Stack(self._session.infra_stack_name)

        self._logger.info(f'**** Infra stack outputs: {json.dumps(stack.outputs)}')

        # for production
        if stack.outputs:
            alb_dns = next((o['OutputValue'] for o in stack.outputs if o['OutputKey'] == 'AlbDns'), None)
        else:
            alb_dns = None

        # for sandboxes
        if not alb_dns:
            # Get ALB arn
            response = self._session.get_cf_client().describe_stack_resource(StackName=self._session.infra_stack_name,
                                                                             LogicalResourceId=Const.MAIN_ALB)
            alb_arn = response['StackResourceDetail']['PhysicalResourceId']
            self._logger.info(f'**** _get_sandbox_public_ip: alb_arn {alb_arn}')

            # Get ALB dns
            response = self._session.get_elb_v2_client().describe_load_balancers(LoadBalancerArns=[alb_arn])
            alb_dns = response['LoadBalancers'][0]['DNSName']

        return alb_dns


class AzureSandboxPublicAddressFetcher(SandboxPublicAddressFetcher):
    def __init__(self, logger: Logger, config: AzureSidecarConfiguration, clients_manager: AzureClientsManager):
        super().__init__(logger=logger, config=config)
        self.clients_manager = clients_manager

    def _get_sandbox_public_ip(self) -> str:
        # Get AG resource group name
        rg_name = self._config.production_id or self._config.sandbox_id

        # Get AG
        if self._config.internet_facing:
            ip = self.clients_manager.network_client.public_ip_addresses.get(resource_group_name=rg_name,
                                                                             public_ip_address_name=Const.AG_PUBLIC_IP)
        else:
            ag = self.clients_manager.network_client.application_gateways.get(resource_group_name=rg_name,
                                                                              application_gateway_name=Const.AG_NAME)
            ip = ag.frontend_ip_configurations.private_ip_address

        if not ip.ip_address:
            raise Exception('ApplicationGateway\'s PublicIpAddress resource does not contain the ip yet')

        return ip.ip_address


class K8sSandboxPublicAddressFetcher(SandboxMetadataFetcher):
    def get_value(self) -> str:
        return ''



import threading
from contextlib import contextmanager
from typing import List, Callable

from sidecar.app_instance_service import StaleAppInstanceException
from sidecar.const import Const, get_app_selector
from sidecar.kub_api_pod_reader import KubApiPodReader
from sidecar.status_maintainer import StatusMaintainer
from sidecar.kub_api_service import IKubApiService
from sidecar.sandbox_error import SandboxError
from logging import Logger
import json


class KubStatusMaintainer(StatusMaintainer):

    def __init__(self, logger: Logger, kub_api_service: IKubApiService):
        super().__init__(logger)
        self._apps_lock = threading.RLock()
        self._services_lock = threading.RLock()
        self.kub_api_service = kub_api_service

    # annotation is like an item in a dynamoDB
    def update_annotation_in_namespace(self, key: str, value: str):
        annotation = {key: value}
        self.kub_api_service.update_namespace(annotation)

    def update_qualiy_status(self, status: str):
        self.update_annotation_in_namespace(Const.QUALIY_STATUS, status)

    def add_sandbox_error(self, error: SandboxError):
        try:
            errors_as_json_str = self.kub_api_service.get_annotation(Const.SANDBOX_ERRORS)
            errors = self._json_str_to_errors(errors_as_json_str)
            errors.append(error)
            errors_as_json_str = self._errors_to_json_str(errors)
            self.update_annotation_in_namespace(Const.SANDBOX_ERRORS, errors_as_json_str)

        except:
            self._logger.exception('Failed to update sandbox errors')

    def update_app_instance_healthcheck_status(self, app_name: str, container_id: str, status: str):
        with self._modify_apps(app_name=app_name, container_id=container_id) as data:
            data['status'] = status

    def update_app_instance_configuration_status(self, app_name: str, container_id: str, status: str):
        with self._modify_apps(app_name=app_name, container_id=container_id) as data:
            data['configuration-status'] = status

    def update_app_instance_artifacts_status(self, app_name: str, container_id: str, status: str):
        with self._modify_apps(app_name=app_name, container_id=container_id) as data:
            data['artifacts-status'] = status

    def update_app_instance_outputs(self, app_name: str, container_id: str, outputs: dict):
        with self._modify_apps(app_name=app_name, container_id=container_id) as data:
            data['outputs'] = outputs

    def add_app_instance_error(self, app_name: str, container_id: str, error: SandboxError):
        with self._modify_apps(app_name=app_name, container_id=container_id) as data:
            data.setdefault('errors', []).append(error.to_dict())

    @contextmanager
    def _modify_apps(self, app_name: str, container_id: str,):
        with self._apps_lock:
            pod_json = self.kub_api_service.try_get_pod_json_by_container_id(container_id=container_id)
            # should update status only for a "live" instance
            if not self._is_pod_live_in_sandbox(pod_json):
                raise StaleAppInstanceException(f"cannot update '{app_name}' since the app "
                                                f"instance is no longer a part of the sandbox. infra id: {container_id}")
            apps_info_json = KubApiPodReader.get_apps_info_json(pod_json)

            yield apps_info_json[app_name]  # executes the code inside the "with self._modify_apps():"

            annotations = {Const.APPS: json.dumps(apps_info_json)}
            self.kub_api_service.update_pod(KubApiPodReader.get_pod_name(pod_json), annotations)

    def update_logical_app_healthcheck_status(self, app_name: str, status: str):
        def healthcheck_status_updater(annotations: dict):
            annotations[Const.HEALTH_CHECK_STATUS] = status

        self._modify_logical_apps(app_name=app_name,
                                  modifier=healthcheck_status_updater)

    def update_logical_app_artifacts_status(self, app_name: str, status: str):
        def artifacts_status_updater(annotations: dict):
            annotations[Const.ARTIFACTS_INTO_SIDECAR_STATUS] = status

        self._modify_logical_apps(app_name=app_name,
                                  modifier=artifacts_status_updater)

    def add_logical_app_error(self, app_name: str, error: SandboxError):
        def add_error_modifier(annotations: dict):
            errors = self._json_str_to_errors(annotations[Const.APP_ERRORS])
            errors.append(error)
            annotations[Const.APP_ERRORS] = self._errors_to_json_str(errors)

        self._modify_logical_apps(app_name=app_name,
                                  modifier=add_error_modifier)

    def _modify_logical_apps(self, app_name: str, modifier: Callable[[dict], None]):
        with self._services_lock:
            # TODO: rethink it !!! not so good to mark both services when only one got checked
            app_services = [service for service in self.kub_api_service.get_all_services()
                            if service['spec']['selector'] == {**{get_app_selector(app_name): app_name}}]

            for app_service in app_services:
                annotations = app_service['metadata']['annotations']

                modifier(annotations)

                self.kub_api_service.update_service(
                    name=app_service['metadata']['name'],
                    data={'metadata': {'annotations': annotations}}
                )

    @staticmethod
    def _is_pod_live_in_sandbox(pod_json) -> bool:
        return True if pod_json and \
                       not KubApiPodReader.is_pod_ended(pod_json) and \
                       not KubApiPodReader.is_pod_terminating(pod_json) else False

    @staticmethod
    def _json_str_to_errors(errors_as_json_str: str) -> List[SandboxError]:
        errors_as_json = json.loads(errors_as_json_str)
        errors = [SandboxError.from_dict(d) for d in errors_as_json or []]
        return errors

    @staticmethod
    def _errors_to_json_str(errors: List[SandboxError]) -> str:
        errors_as_json = [error.to_dict() for error in errors]
        errors_as_json_str = json.dumps(errors_as_json)
        return errors_as_json_str




from abc import ABCMeta, abstractmethod
from logging import Logger
from typing import List
from sidecar.sandbox_error import SandboxError


class StatusMaintainer:
    __metaclass__ = ABCMeta

    def __init__(self, logger: Logger):
        self._logger = logger

    @abstractmethod
    def update_qualiy_status(self, status: str):
        raise NotImplementedError

    @abstractmethod
    def add_sandbox_error(self, error: SandboxError):
        raise NotImplementedError


class IngressRouteRequest:
    def __init__(self, listener_port: int, path: str, host: str, app_port: int, app_name: str, color: str):
        self.listener_port = listener_port
        self.path = path
        self.host = host
        self.app_port = app_port
        self.app_name = app_name
        self.color = color